function [b r LAMBDA stats result] = epistasis(MODEL_NAME, X, y, LAMBDA, MAX, CONVERGE_THRESHOLD, DISP)
%[b r LAMBDA stats result] = epistasis(MODEL_NAME, X, y, LAMBDA, MAX,
%CONVERGE_THRESHOLD, DISP) performs epistasis study. 
% MODEL_NAME: 'MULTIPLICATIVE WB' for Model I, 'MULTIPLICATIVE WB
% THRESHOLD' for Model II, MULTIPLICATIVE for Model III, and 'THRESHOLD'
% for Model IV.
% X: genetic data (n * 2)
% y: gene expression (n * 1)
% LAMBDA: hyperparameter for regularization (default: 1e-1)
% MAX: max iteration parameter (default: 1e+5)
% CONVERGE_THRESHOLD: a threshold for converge (default: 1e-3)
% DISP: boolean if it displays information at an iteration
%
% created by Mingon Kang (mingon.kang@mavs.uta.edu) in 2014

if nargin < 4, LAMBDA = 1e-1; MAX = 1e+5; CONVERGE_THRESHOLD = 1e-3; DISP = true; end
if nargin < 5, MAX = 1e+5; CONVERGE_THRESHOLD = 1e-3; DISP = true;   end
if nargin < 6, CONVERGE_THRESHOLD = 1e-3; DISP = true; end
if nargin < 7, DISP = true; end

try

% init
b0 = 0;
b_a = 1;
b_b = 1;
b_ab = 1;
r_a = 1;
r_b = 1;

% calculate the number of 3*3 table
W = ones(size(y));
for i = 0 : 2
    for j = 0 : 2
        h = (X(:, 1) == i & X(:, 2) == j);
        w = sum(h);
        if w == 0, w = 1; end
        W(h) = 1 / w;
    end
end

K = InterPower(MODEL_NAME, X);

p_param = [b_a b_b b_ab r_a r_b b0]';  % previous param
for iter = 1 : MAX
    
    R = y - b_a .* X(:, 1) - b_b .* X(:, 2);
    C = [-R .* b_ab .* (r_b .^ K(:, 2)) .* K(:, 1); b_ab ^ 2 .* (r_b .^ (2 * K(:, 2))) .* K(:, 1)];
    power = [K(:, 1) - 1; 2 * K(:, 1) - 1];
    max_power = max(power);
    coef_power = zeros(max_power + 1, 1);
    for i = 0 : max_power
        coef_power(max_power - i + 1) = sum(C(power == i));
    end
    
    r1 = REAL(roots(coef_power));
    r_a = r1(r1 > 0);
    if isempty(r_a), r_a = 1; end
    r_a = r_a(1);

    C = [-R .* b_ab .* (r_a .^ K(:, 1)) .* K(:, 2); b_ab ^ 2 .* (r_a .^ (2 * K(:, 1))) .* K(:, 2)];
    power = [K(:, 2) - 1; 2 * K(:, 2) - 1];
    max_power = max(power);
    coef_power = zeros(max_power + 1, 1);
    for i = 0 : max_power
        coef_power(max_power - i + 1) = sum(C(power == i));
    end
    r2 = REAL(roots(coef_power));
    r_b = r2(r2 > 0);
    if isempty(r_b), r_b = 1; end
    r_b = r_b(1);
    
    b_ab = nansum(W .* (r_a .^ K(:, 1) .* r_b .^ K(:, 2) .* (y - b_a * X(:, 1) - b_b * X(:, 2)))) / nansum(W .* (r_a.^K(:, 1) .* r_b.^K(:, 2)).^2 + LAMBDA);
    b_a = nansum(W .* (y .* X(:, 1) - b_b * X(:, 1) .* X(:, 2) - b_ab .* r_a .^ K(:, 1) .* r_b .^ K(:, 2) .* X(:, 1))) / nansum(W .* X(:, 1) .^ 2 + LAMBDA);
    b_b = nansum(W .* (y .* X(:, 2) - b_b * X(:, 1) .* X(:, 2) - b_ab .* r_a .^ K(:, 1) .* r_b .^ K(:, 2) .* X(:, 2))) / nansum(W .* X(:, 2) .^ 2 + LAMBDA);  
    
    param = [b_a b_b b_ab r_a r_b b0]';
    converge = norm(abs(param - p_param));
    
    if DISP,
        fprintf('SQ_ERR: %d, CONV: %d \n', ...
            sum((y - b_a * X(:, 1) - b_b * X(:, 2) - b_ab .* (r_a .^ K(:, 1)) .* (r_b .^ K(:, 2))).^2), ...
            converge);
    end
    if converge < CONVERGE_THRESHOLD
        if DISP, fprintf('Converged at %d\n', iter); end
        break;
    end
    p_param = param;
end
if iter == MAX
    result = 0; 
    b = zeros(4, 1);
    r = zeros(2, 1);
    stats = [];
    if DISP, disp('exceed the maximum iteration'); end
else
    result = 1;
    b0 = mean(y) - mean([X (param(4).^K(:, 1)) .* (param(5) .^ K(:, 2))]) * param(1:3);
    b = [b0; b_a; b_b; b_ab];
    r = [r_a; r_b];
    
    XN = [X param(4).^K(:, 1) .* param(5) .^ K(:, 2)];
    stats = regstats(y, XN);
end

catch err
    err.message
    b = [];
    r = [];
    stats = [];
    result = 0;
    return
end
